/* Job.cc
 *   Copyright (C) 2001-2002 Valentin Ziegler and Ren Rebe
 *
 ***************************************************************************
 *                                                                         *
 *   This code is free software; you can redistribute it and / or modify   *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Threads.hh"

#include <unistd.h>
#include <iostream>
#include "Commands.hh"

using namespace std;
using namespace SigC;
using namespace GSMX_utils::Threads;
using namespace GSMX_utils::Threads::Private;

/************
 * Command0 *
 ************/

Command0 :: Command0 ()
{
  call = 0;
  target = 0;
}

void Command0 :: Recaller :: Execute ()
{
  ((references -> target) ->* (references -> call)) ();
}

void Command0 :: operator () ()
{
  if (!call)
    {   
      ExceptUnbound e;
      throw e;
    }
  DropCmd (new Command0 :: Recaller (this));
}

/***********
 * Funct__ *
 ***********/

Funct__::Funct__ () 
{ 
  // Does nothing, bot Funct__ needs a protected ctor ! 
}

void Funct__::Bind (Job* j)
{
  target = j;
}

void Funct__::Hold ()
{
  synchronisation = new Private::CrossExec ();
  target -> PushCmd (synchronisation);
  synchronisation -> Wait ();
}

void Funct__::Release ()
{
  synchronisation -> Release ();
}

/*******
 * Job *
 *******/

Job::Job ()
{
  /* TODO: exceptions */
  wait_after_loop = false;
  waiting = false;
  exit = false;
  ctor.lock ();
  this -> start (NULL);
  ctor.lock (); // Wait for Handshake "Thread is running"
}

Job::~Job ()
{
  exit = true;
  Continue ();
  
  pthread_join (thread_, NULL);
  cout << "THREAD-DEBUG: DTor finished ..." << endl;
}

bool Job::IsRunning () 
{ 
  MLock lazy (waiting__);
  return !waiting;
}

void Job::Ready (bool run_now) 
{
  start_waiting = !run_now;
  on_request.unlock (); // signal "Constructors finished"

  if (!run_now)
    {
      while (start_waiting) // if run_now = false...
	usleep (1000);      // ... wait until Thread entered Wait__ ()

      waiting__.lock ();    // wait until Thread sleeps on Condition
      waiting__.unlock ();
    }
}

void Job::SyncWait ()
{
  Private::CrossExec* synchronisation = new Private::CrossExec ();
  PushCmd (synchronisation);
  synchronisation -> Wait ();
  Wait ();
  synchronisation -> Release ();
}

void Job::AsyncWait ()
{
  PushCmd (new Private::CMDWait (this));
}

bool Job::IsRequest ()
{
  MLock lazy (commands__);
  return (!commands.empty ());
}

void Job::Wait__ ()
{
  waiting__.lock ();
  waiting = true;
  start_waiting = false;
  new_request.wait (waiting__);
  waiting = false;
  waiting__.unlock ();
}

void Job::Wait ()
{
  wait_after_loop = true;
}

void Job::Continue ()
{
  waiting__.lock ();
  wait_after_loop = false;
  // This makes sure Continue () works when called outside second thread
  if (waiting)
    new_request.signal ();
  waiting__.unlock ();
}

void* Job::main (void* data)
{
  on_request.lock ();
  ctor.unlock (); // let the ctor finish
  on_request.lock (); // wait for Ready () call from iheritenced class
  on_request.unlock ();
  
  if (start_waiting) Wait__ ();
  
  // main loop
  while (!exit)
    {
      Private::Command* command = 0;
      
      commands__.lock ();
      if (!commands.empty ())
	{
	  command = commands.front ();
	  commands.pop ();
	}	
      commands__.unlock ();

      if (command)
	{
	  command -> Execute ();
	  delete command;
	}      
      else
	{
	  if (wait_after_loop) 
	    Wait__ ();
	  else
	    Loop ();
	}    
    }
  
  cout << "THREAD-DEBUG: main finished ..." << endl;
}

void Job::PushCmd (Private::Command* cmd)
{
  int pending_commands;

  if (exit)
    {
      ExceptTerminated e;
      throw e;
    }

  do
    {
      MLock lazy (commands__);
      pending_commands = commands.size ();
      if (pending_commands < THREAD_MAX_JOB_CMDS)
	commands.push (cmd);
      else // command queue is flooded ! 
	usleep (10000);
    } while (pending_commands >= THREAD_MAX_JOB_CMDS);
  
  new_request.signal ();
}

CrossExec::CrossExec ()
{
  start.lock ();
  finished.lock ();
}

void CrossExec::Execute ()
{
  start.unlock ();
  finished.lock ();
}


//// Former sigc++ stuff addapted from code by Karl Nelson ////

Mutex::Mutex(const MutexAttr attr)
{
  pthread_mutex_init(&mutex_,attr.impl_);
}

// (needs work)
Mutex::~Mutex()
{
  destroy();
}

int Mutex::lock()    {return pthread_mutex_lock(&mutex_);}
int Mutex::trylock() {return pthread_mutex_trylock(&mutex_);}
int Mutex::unlock()  {return pthread_mutex_unlock(&mutex_);}
int Mutex::destroy() {return pthread_mutex_destroy(&mutex_);}


Condition::Condition(const CondAttr &attr)
{
  pthread_cond_init(&cond_,attr.impl_);
}

Condition::~Condition()
{
  destroy();
}

int Condition::signal()       {return pthread_cond_signal(&cond_);}
int Condition::broadcast()    {return pthread_cond_broadcast(&cond_);}
int Condition::wait(Mutex &m) {return pthread_cond_wait(&cond_,m);}
int Condition::wait(Mutex &m,struct timespec* spec)
{
  return pthread_cond_timedwait(&cond_,m,spec);
}
int Condition::destroy()      {return pthread_cond_destroy(&cond_);}


void Semaphore::up()
{
  access_.lock();
  value_++;
  access_.unlock();
  sig_.signal();
}

void Semaphore::down()
{
  access_.lock();
  while (value_<1)
    {sig_.wait(access_);}
  value_--;
  access_.unlock();
}

Semaphore::Semaphore(int value):value_(value) {}
Semaphore::~Semaphore() {}

void* Thread::call_main_(void* obj)
{
  Thread *thread=(Thread*)obj;
  return thread->main(thread->arg_);
}

Thread::Thread(const ThreadAttr &attr):attr_(attr) {}
Thread::~Thread() {}

int Thread::detach()
{
  return pthread_detach(thread_);
}

int Thread::start(void* arg)
{
  arg_=arg;
  Thread *t=this;
  return pthread_create(&thread_,attr_.impl_,call_main_,t);
}

void* Private___::get()
{
  void* value;
  value=(void*)(pthread_getspecific(key_));
  return value;
}

void Private___::set(void *value)
{
  pthread_setspecific(key_,value);
}

void Private___::create(void (*dtor)(void*))
{
  pthread_key_create(&key_,dtor);
}

void Private___::destroy()
{
  pthread_key_delete(key_);
}

MutexAttr Mutex::Default={0};
CondAttr Condition::Default={0};
ThreadAttr Thread::Default={0};

